# coding: utf-8
# file generated by setuptools_scm
# don't change, don't track in version control
version = '0.1.5.post2.dev2+gb159f13.d20210528'
version_tuple = (0, 1, 5, 'post2', 'dev2+gb159f13', 'd20210528')



import sys
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

from .tools import *

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # < Python 3.8: Use backport module
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
    del version
except PackageNotFoundError:
    pass



from .Base import Base
from demuxEM.pipeline import run_pipeline


class DemuxEM(Base):
    """
Run the DemuxEM pipeline for cell-hashing/nucleus-hashing data.

Usage:
  demuxEM [options] <input_raw_gene_bc_matrices_h5> <input_hto_csv_file> <output_name>
  demuxEM -h | --help
  demuxEM -v | --version

Arguments:
  input_raw_gene_bc_matrices_h5           Input raw RNA expression matrix in 10x hdf5 format.
  input_hto_csv_file                      Input HTO (antibody tag) count matrix in CSV format, with antibody names being rows and cell barcodes being columns. NOTICE: The first column (i.e. antibody names) MUST have a header name, usually "Antibody", to get correct demultiplexing results.
  output_name                             Output name. All outputs will use it as the prefix.

Options:
  -p <number>, --threads <number>            Number of threads. [default: 1]
  --genome <genome>                          Reference genome name. If not provided, we will infer it from the expression matrix file.
  --alpha-on-samples <alpha>                 The Dirichlet prior concentration parameter (alpha) on samples. An alpha value < 1.0 will make the prior sparse. [default: 0.0]

  --min-num-genes <number>                   We only demultiplex cells/nuclei with at least <number> of expressed genes. [default: 100]
  --min-num-umis <number>                    We only demultiplex cells/nuclei with at least <number> of UMIs. [default: 100]
  --min-signal-hashtag <count>               Any cell/nucleus with less than <count> hashtags from the signal will be marked as unknown. [default: 10.0]
  --random-state <seed>                      The random seed used in the KMeans algorithm to separate empty ADT droplets from others. [default: 0]

  --generate-diagnostic-plots                Generate a series of diagnostic plots, including the background/signal between HTO counts, estimated background probabilities, HTO distributions of cells and non-cells etc.
  --generate-gender-plot <genes>             Generate violin plots using gender-specific genes (e.g. Xist). <gene> is a comma-separated list of gene names.

  -v, --version                              Show version.
  -h, --help                                 Print out help information.

Outputs:
  output_name_demux.zarr.zip                             RNA expression matrix with demultiplexed sample identities in Zarr format.
  output_name.out.demuxEM.zarr.zip                       DemuxEM-calculated results in Zarr format, containing two datasets, one for HTO and one for RNA.
  output_name.ambient_hashtag.hist.pdf                   Optional output. A histogram plot depicting hashtag distributions of empty droplets and non-empty droplets.
  output_name.background_probabilities.bar.pdf           Optional output. A bar plot visualizing the estimated hashtag background probability distribution.
  output_name.real_content.hist.pdf                      Optional output. A histogram plot depicting hashtag distributions of not-real-cells and real-cells as defined by total number of expressed genes in the RNA assay.
  output_name.rna_demux.hist.pdf                         Optional output. This figure consists of two plots. The first one is a horizontal bar plot depicting the percentage of RNA barcodes with at least one HTO count. The second plot is a histogram plot depicting RNA UMI distribution for singlets, doublets and unknown cells.
  output_name.gene_name.violin.pdf                       Optional outputs. Violin plots depicting gender-specific gene expression across samples. We can have multiple plots if a gene list is provided in '--generate-gender-plot' option.

Examples:
  demuxEM -p 8 --generate-diagnostic-plots sample_raw_gene_bc_matrices.h5 sample_hto.csv sample_output
    """

    def execute(self):
        kwargs = {
            "n_jobs": int(self.args["--threads"]),
            "genome": self.args["--genome"],
            "alpha": float(self.args["--alpha-on-samples"]),
            "min_num_genes": int(self.args["--min-num-genes"]),
            "min_num_umis": int(self.args["--min-num-umis"]),
            "min_signal": float(self.args["--min-signal-hashtag"]),
            "random_state": int(self.args["--random-state"]),
            "gen_plots": self.args["--generate-diagnostic-plots"],
            "gen_gender_plot": self.split_string(self.args["--generate-gender-plot"]),
        }

        run_pipeline(
            self.args["<input_raw_gene_bc_matrices_h5>"],
            self.args["<input_hto_csv_file>"],
            self.args["<output_name>"],
            **kwargs
        )



import sys
from docopt import docopt
from . import __version__ as VERSION
from .commands import demuxEM


def main():
    command = demuxEM(sys.argv[1:], version = VERSION)
    command.execute()


if __name__ == "__main__":
    main()



from docopt import docopt
from typing import List


class Base:
    """Base class for the commands"""

    def __init__(self, command_args: List[str], version=None):
        self.args = docopt(self.__doc__, argv=command_args, version=version)

    def split_string(self, astring: str, sep: str = ",") -> List[str]:
        return astring.split(sep) if astring is not None else []

    def convert_to_int(self, value: str) -> int:
        return int(value) if value is not None else None

    def convert_to_float(self, value: str) -> float:
    	return float(value) if value is not None else None

    def execute(self):
        raise NotImplementedError



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from natsort import natsorted

from typing import Tuple, List
from pegasusio import UnimodalData



def plot_hto_hist(hashing_data: UnimodalData, attr: str, out_file: str, alpha: float = 0.5, dpi: int = 500, figsize: Tuple[float, float] = None) -> None:
    idx_signal = np.isin(hashing_data.obs[attr], "signal")
    signal = hashing_data.obs.loc[idx_signal, "counts"]
    background = hashing_data.obs.loc[~idx_signal, "counts"]
    bins = np.logspace(0, np.log10(max(signal.max(), background.max())), 501)
    plt.hist(background, bins, alpha=alpha, label="background", log=True)
    plt.hist(signal, bins, alpha=alpha, label="signal", log=True)
    plt.legend(loc="upper right")
    ax = plt.gca()
    ax.set_xscale("log")
    ax.set_xlabel("Number of hashtag UMIs (log10 scale)")
    ax.set_ylabel("Number of cellular barcodes (log10 scale)")
    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)
    plt.savefig(out_file, dpi=dpi)
    plt.close()


def plot_rna_hist(
    rna_data: UnimodalData, hashing_data: UnimodalData, out_file: str, plot_attr: str = "n_counts", cat_attr: str = "demux_type", dpi: int = 500, figsize: Tuple[float, float] = None
) -> None:
    fig, axes = plt.subplots(2, 1, gridspec_kw={'height_ratios': [1, 9], 'hspace': 0.6}, figsize = figsize, dpi = dpi)

    # Percentage of RNA barcodes having HTO tags
    nhto = rna_data.obs_names.isin(hashing_data.obs_names).sum()
    total = rna_data.shape[0]

    ax = axes[0]
    p1 = nhto * 100.0 / total
    p2 = (total - nhto) * 100.0 / total
    labels = ['Has HTO', 'No HTO']
    ax.barh(y = 0, width = p1, color = 'red', height = 0.5, label = labels[0])
    ax.barh(y = 0, width = p2, left = p1, color = 'gray', height = 0.5, label = labels[1])
    ax.set_yticks([])
    ax.set_ylabel('RNA    \nbarcodes', rotation = 0, ha = 'right', va = 'center')
    ax.set_xlim(left = 0, right = 100)
    ax.set_xlabel('Percentage')
    ax.legend(ncol=2, bbox_to_anchor=(1, 1), loc='lower right', fontsize = 'small')

    # RNA histogram
    bins = np.logspace(
        np.log10(min(rna_data.obs[plot_attr])), np.log10(max(rna_data.obs[plot_attr])), 101
    )
    cat_vec = rna_data.obs[cat_attr]
    ax = axes[1]
    ax.hist(
        rna_data.obs.loc[np.isin(cat_vec, "singlet"), plot_attr],
        bins,
        alpha=0.5,
        label="singlet",
    )
    ax.hist(
        rna_data.obs.loc[np.isin(cat_vec, "doublet"), plot_attr],
        bins,
        alpha=0.5,
        label="doublet",
    )
    ax.hist(
        rna_data.obs.loc[np.isin(cat_vec, "unknown"), plot_attr],
        bins,
        alpha=0.5,
        label="unknown",
    )
    ax.legend(loc="upper right")
    ax.set_xscale("log")
    ax.set_xlabel("Number of RNA UMIs (log10 scale)")
    ax.set_ylabel("Number of cellular barcodes")

    fig.savefig(out_file)


def plot_bar(heights: List[float], tick_labels: List[str], xlabel: str, ylabel: str, out_file: str, dpi: int = 500, figsize: Tuple[float, float] = None) -> None:
    plt.bar(
        x=np.linspace(0.5, heights.size - 0.5, heights.size),
        height=heights,
        tick_label=tick_labels,
    )
    ax = plt.gca()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)
    rotation = 90 if max([len(x) for x in tick_labels]) > 6 else 0
    plt.tick_params(axis="x", labelsize=7, labelrotation=rotation)
    plt.tight_layout()
    plt.savefig(out_file, dpi=dpi)
    plt.close()


def plot_gene_violin(
    data: UnimodalData,
    gene_name: str,
    out_file: str,
    title: str = None,
    dpi: int = 500,
    figsize: Tuple[float, float] = None,
    linewidth: float = None,
    inner: str = "box",
) -> None:
    df = pd.DataFrame(
        data[:, gene_name].X.toarray(),
        index=data.obs_names,
        columns=[gene_name],
    )
    df["assignment"] = data.obs["demux_type"].astype(str)
    idx_singlet = np.isin(data.obs["demux_type"], "singlet")
    singlets = data.obs.loc[idx_singlet, "assignment"].astype(str)
    df.loc[idx_singlet, "assignment"] = singlets
    categories = natsorted(singlets.unique())
    categories.extend(["doublet", "unknown"])
    df["assignment"] = pd.Categorical(df["assignment"], categories=categories)
    xlabel = "assignment"
    ylabel = gene_name

    sns.violinplot(
        x=xlabel, y=ylabel, data=df, linewidth=linewidth, cut=0, inner=inner
    )

    ax = plt.gca()
    ax.grid(False)
    ax.set_ylabel("log(TP100K+1)")
    if title is not None:
        ax.set_title(title)

    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)

    rotation = 90 if max([len(x) for x in df[xlabel].unique()]) > 6 else 0
    plt.tick_params(axis="x", labelsize=7, labelrotation=rotation)
    plt.tight_layout()
    plt.savefig(out_file, dpi=dpi)
    plt.close()


def plot_heatmap(vec1: List[int], vec2: List[int], out_file: str, dpi: int = 500, xlabel: str = "", ylabel: str = "") -> None:
    df = pd.crosstab(vec1, vec2)
    df.columns.name = df.index.name = ""

    ax = plt.gca()
    ax.xaxis.tick_top()
    ax = sns.heatmap(df, annot=True, fmt="d", cmap="inferno", ax=ax)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    plt.tight_layout()
    plt.savefig(out_file, dpi=500)
    plt.close()


def plot_dataframe_bar(df: pd.DataFrame, ylabel: str, out_file: str, dpi: int = 500, figsize: Tuple[float, float] = None) -> None:
    if df.shape[1] == 1:
        df.plot.bar(legend=False)
    else:
        df.plot.bar()
    ax = plt.gca()
    ax.set_ylabel(ylabel)
    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)
    plt.savefig(out_file, dpi=dpi)
    plt.close()



import numpy as np
import matplotlib.pyplot as plt
from typing import List, Tuple

from pegasusio import read_input, UnimodalData
from . import estimate_background_probs, demultiplex


def down_sampling(rna_gt: UnimodalData, hto_gt: UnimodalData, probs: List[float], n_threads: int = 1):
    f = np.vectorize(lambda x, p: np.random.binomial(int(x + 1e-4), p, size=1)[0])

    nsample = rna_gt.shape[0]
    nhto = hto_gt.X.sum()

    fracs = []
    accuracy = []
    for p in probs:
        rna_data = rna_gt.copy()
        hto_data = hto_gt.copy()

        hto_data.X.data = f(hto_data.X.data, p)
        idx = hto_data.X.sum(axis=1).A1 > 0
        hto_data = hto_data[idx,].copy(deep = False)
        fracs.append(hto_data.X.sum() / nhto)

        estimate_background_probs(hto_data)
        demultiplex(rna_data, hto_data, n_threads=n_threads)
        accuracy.append(
            sum(
                rna_data.obs["assignment"].values.astype("str")
                == rna_gt.obs["assignment"].values.astype("str")
            )
            / nsample
        )

    return fracs, accuracy


def plot_down_sampling(
    demuxEM_res_file: str, 
    out_file: str,
    probs: List[float] = [i / 10.0 for i in range(9, 0, -1)],
    n_threads: int = 1,
    dpi: int = 500,
    figsize: Tuple[float, float] = None,
):
    data = read_input(demuxEM_res_file)
    rna_gt = data.get_data(modality = "rna")
    hto_gt = data.get_data(modality = "hashing")

    fracs, accuracy = down_sampling(rna_gt, hto_gt, probs, n_threads=n_threads)

    plt.plot(fracs, accuracy, ".-")
    ax = plt.gca()
    ax.set_xlim(1.0, 0.0)
    ax.set_ylim(0.79, 1.01)
    vals = ax.get_yticks()
    ax.set_yticklabels(["{:.0%}".format(v) for v in vals])
    ax.set_xlabel("Fraction of hashtag UMIs")
    ax.set_ylabel("Consistency")
    if figsize is not None:
        plt.gcf().set_size_inches(*figsize)
    plt.savefig(out_file, dpi=dpi)
    plt.close()



from .demuxEM import estimate_background_probs, demultiplex, attach_demux_results
from .plot import (
    plot_hto_hist,
    plot_rna_hist,
    plot_bar,
    plot_gene_violin,
    plot_heatmap,
    plot_dataframe_bar,
)
from .down_sampling import plot_down_sampling



import numpy as np
import pegasusio as io

import logging
logger = logging.getLogger(__name__)

from demuxEM.tools import *



def run_pipeline(input_rna_file, input_hto_file, output_name, **kwargs):
    # load input rna data
    data = io.read_input(input_rna_file, genome=kwargs["genome"], modality="rna")
    data.subset_data(modality_subset=['rna'])
    data.concat_data() # in case of multi-organism mixing data
    genome = data.uns["genome"]

    # load input hashing data
    data.update(io.read_input(input_hto_file, genome=genome, modality="hashing"))

    # Extract rna and hashing data
    rna_data = data.get_data(modality="rna")
    hashing_data = data.get_data(modality="hashing")

    # Filter the RNA matrix
    rna_data.obs["n_genes"] = rna_data.X.getnnz(axis=1)
    rna_data.obs["n_counts"] = rna_data.X.sum(axis=1).A1
    obs_index = np.logical_and.reduce(
        (
            rna_data.obs["n_genes"] >= kwargs["min_num_genes"],
            rna_data.obs["n_counts"] >= kwargs["min_num_umis"],
        )
    )
    rna_data._inplace_subset_obs(obs_index)

    # run demuxEM
    estimate_background_probs(hashing_data, random_state=kwargs["random_state"])

    demultiplex(
        rna_data,
        hashing_data,
        min_signal=kwargs["min_signal"],
        alpha=kwargs["alpha"],
        n_threads=kwargs["n_jobs"],
    )

    # annotate raw matrix with demuxEM results
    demux_results = attach_demux_results(input_rna_file, rna_data)

    # generate plots
    if kwargs["gen_plots"]:
        plot_hto_hist(
            hashing_data, "hto_type", output_name + ".ambient_hashtag.hist.pdf", alpha=1.0
        )
        plot_bar(
            hashing_data.uns["background_probs"],
            hashing_data.var_names,
            "Sample ID",
            "Background probability",
            output_name + ".background_probabilities.bar.pdf",
        )
        plot_hto_hist(
            hashing_data, "rna_type", output_name + ".real_content.hist.pdf", alpha=0.5
        )
        plot_rna_hist(rna_data, hashing_data, output_name + ".rna_demux.hist.pdf")
        logger.info("Diagnostic plots are generated.")

    if len(kwargs["gen_gender_plot"]) > 0:
        rna_data.matrices["raw.X"] = rna_data.X.copy()
        rna_data.as_float()
        scale = 1e5 / rna_data.X.sum(axis=1).A1
        rna_data.X.data *= np.repeat(scale, np.diff(data.X.indptr))
        rna_data.X.data = np.log1p(rna_data.X.data)

        for gene_name in kwargs["gen_gender_plot"]:
            plot_gene_violin(
                rna_data,
                gene_name,
                "{output_name}.{gene_name}.violin.pdf".format(
                    output_name=output_name, gene_name=gene_name
                ),
                title="{gene_name}: a gender-specific gene".format(gene_name=gene_name),
            )

        logger.info("Gender-specific gene expression violin plots are generated.")

    # output results
    io.write_output(demux_results, output_name + "_demux.zarr.zip")
    io.write_output(data, output_name + ".out.demuxEM.zarr.zip")

    # output summary statistics
    print("\nSummary statistics:")
    print("total\t{}".format(rna_data.shape[0]))
    for name, value in rna_data.obs["demux_type"].value_counts().iteritems():
        print("{}\t{}".format(name, value))



import numpy as np
import pandas as pd
import time
from natsort import natsorted

import multiprocessing
from sklearn.cluster import KMeans

from typing import List, Tuple
from pegasusio import UnimodalData, MultimodalData, read_input

import logging
logger = logging.getLogger(__name__)


def estimate_background_probs(hashing_data: UnimodalData, random_state: int = 0) -> None:
    """For cell-hashing data, estimate antibody background probability using KMeans algorithm.

    Parameters
    ----------
    hashing_data: ``UnimodalData``
        Annotated data matrix for antibody.

    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.

    Returns
    -------
    ``None``

    Update ``hashing_data.uns``:
        * ``hashing_data.uns["background_probs"]``: estimated antibody background probability.

    Example
    -------
    >>> estimate_background_probs(hashing_data)
    """
    hashing_data.obs["counts"] = hashing_data.X.sum(axis=1).A1
    counts_log10 = np.log10(hashing_data.obs["counts"].values.reshape(-1, 1))
    kmeans = KMeans(n_clusters=2, random_state=random_state).fit(counts_log10)
    signal = 0 if kmeans.cluster_centers_[0] > kmeans.cluster_centers_[1] else 1
    hashing_data.obs["hto_type"] = "background"
    hashing_data.obs.loc[kmeans.labels_ == signal, "hto_type"] = "signal"

    idx = np.isin(hashing_data.obs["hto_type"], "background")
    pvec = hashing_data.X[idx,].sum(axis=0).A1
    back_probs = pvec / pvec.sum()

    idx = back_probs <= 0.0
    if idx.sum() > 0:
        logger.warning(f"Detected {idx.sum()} antibody barcodes {','.join(hashing_data.var_names[idx])} with 0 counts in the background! These barcodes are likely not in the experiment and thus removed.")
        hashing_data._inplace_subset_var(~idx)
        back_probs = back_probs[~idx]

    hashing_data.uns["background_probs"] = back_probs
    logger.info("Background probability distribution is estimated.")


def estimate_probs(arr: List[float], pvec: List[float], alpha: float, alpha_noise: float, tol: float) -> List[float]:
    probs = np.zeros(pvec.size + 1)
    old_probs = np.zeros(pvec.size + 1)
    z = np.zeros(pvec.size + 1)
    noise = pvec.size

    # Estimate MLE without Generalized Dirichlet prior
    probs_mle = arr / arr.sum()
    probs[noise] = (probs_mle / pvec).min() + 0.01
    probs[:-1] = np.maximum(probs_mle - probs[noise] * pvec, 0.01)
    probs = probs / probs.sum()

    # EM algorithm
    i = 0
    eps = 1.0
    while eps > tol:
        i += 1
        old_probs[:] = probs[:]
        # E step
        z[:-1] = alpha - 1.0
        z[noise] = alpha_noise - 1.0
        for j in range(pvec.size):
            if arr[j] > 0:
                p = probs[j] / (probs[noise] * pvec[j] + probs[j])
                z[j] += arr[j] * p
                z[noise] += arr[j] * (1.0 - p)
        # M step
        idx = z > 0.0
        probs[idx] = z[idx] / z[idx].sum()
        probs[~idx] = 0.0
        eps = np.linalg.norm(probs - old_probs, ord=1)
        # print ("i = {}, eps = {:.2g}.".format(i, eps))

    return probs


def get_droplet_info(probs: List[float], sample_names: List[str]) -> Tuple[str, str]:
    ids = np.nonzero(probs >= 0.1)[0]
    ids = ids[np.argsort(probs[ids])[::-1]]
    return (
        "singlet" if ids.size == 1 else "doublet",
        ",".join([sample_names[i] for i in ids]),
    )


def calc_demux(rna_data: UnimodalData, hashing_data: UnimodalData, nsample: int, min_signal: float, probs: str = "raw_probs") -> None:
    demux_type = np.full(rna_data.shape[0], "unknown", dtype="object")
    assignments = np.full(rna_data.shape[0], "", dtype="object")

    signals = hashing_data.obs["counts"].reindex(rna_data.obs_names, fill_value=0.0).values * (
        1.0 - rna_data.obsm[probs][:, nsample]
    )
    idx = signals >= min_signal

    tmp = rna_data.obsm[probs][idx,]
    norm_probs = tmp[:, 0:nsample] / (1.0 - tmp[:, nsample])[:, None]

    values1 = []
    values2 = []
    for i in range(norm_probs.shape[0]):
        droplet_type, droplet_id = get_droplet_info(norm_probs[i,], hashing_data.var_names)
        values1.append(droplet_type)
        values2.append(droplet_id)

    demux_type[idx] = values1
    rna_data.obs["demux_type"] = pd.Categorical(
        demux_type, categories=["singlet", "doublet", "unknown"]
    )
    assignments[idx] = values2
    rna_data.obs["assignment"] = pd.Categorical(
        assignments, categories=natsorted(np.unique(assignments))
    )


def has_duplicate_names(names: List[str]) -> bool:
    for name in names:
        if name.find(".#~") >= 0:
            return True
    return False


def remove_suffix(assigns: List[str]) -> pd.Categorical:
    assigns = assigns.astype(str)
    results = []
    for value in assigns:
        fields = value.split(",")
        for i, item in enumerate(fields):
            pos = item.find(".#~")
            if pos >= 0:
                fields[i] = item[:pos]
        results.append(",".join(fields))
    results = np.array(results)
    return pd.Categorical(results, categories=natsorted(np.unique(results)))


def demultiplex(
    rna_data: UnimodalData,
    hashing_data: UnimodalData,
    min_signal: float = 10.0,
    alpha: float = 0.0,
    alpha_noise: float = 1.0,
    tol: float = 1e-6,
    n_threads: int = 1,
):
    """Demultiplexing cell/nucleus-hashing data, using the estimated antibody background probability calculated in ``demuxEM.estimate_background_probs``.

    Parameters
    ----------
    rna_data: ``UnimodalData``
        Data matrix for gene expression matrix.

    hashing_data: ``UnimodalData``
        Data matrix for HTO count matrix.

    min_signal: ``float``, optional, default: ``10.0``
        Any cell/nucleus with less than ``min_signal`` hashtags from the signal will be marked as ``unknown``.

    alpha: ``float``, optional, default: ``0.0``
        The Dirichlet prior concentration parameter (alpha) on samples. An alpha value < 1.0 will make the prior sparse.

    alpha_noise: ``float``, optional, default: ``1.0``
        The Dirichlet prior concenration parameter on the background noise.

    tol: ``float``, optional, default: ``1e-6``
        Threshold used for the EM convergence.

    n_threads: ``int``, optional, default: ``1``
        Number of threads to use. Must be a positive integer.

    Returns
    -------
    ``None``

    Update ``data.obs``:
        * ``data.obs["demux_type"]``: Demultiplexed types of the cells. Either ``singlet``, ``doublet``, or ``unknown``.
        * ``data.obs["assignment"]``: Assigned samples of origin for each cell barcode.
        * ``data.obs["assignment.dedup"]``: Only exist if one sample name can correspond to multiple feature barcodes. In this case, each feature barcode is assigned a unique sample name.

    Examples
    --------
    >>> demultiplex(rna_data, hashing_data)
    """
    nsample = hashing_data.shape[1]
    rna_data.uns["background_probs"] = hashing_data.uns["background_probs"]
    assert (rna_data.uns["background_probs"] <= 0.0).sum() == 0

    idx_df = rna_data.obs_names.isin(hashing_data.obs_names)
    hashing_data.obs["rna_type"] = "background"
    hashing_data.obs.loc[rna_data.obs_names[idx_df], "rna_type"] = "signal"

    if nsample == 1:
        logger.warning("Detected only one barcode, no need to demultiplex!")
        rna_data.obsm["raw_probs"] = np.zeros((rna_data.shape[0], nsample + 1))
        rna_data.obsm["raw_probs"][:, 0] = 1.0
        rna_data.obsm["raw_probs"][:, 1] = 0.0
        rna_data.obs["demux_type"] = "singlet"
        rna_data.obs["assignment"] = hashing_data.var_names[0]
    else:
        if nsample == 2:
            logger.warning("Detected only two barcodes, demultiplexing accuracy might be affected!")

        ncalc = idx_df.sum()
        if ncalc < rna_data.shape[0]:
            nzero = rna_data.shape[0] - ncalc
            logger.warning(
                "Warning: {} cells do not have ADTs, percentage = {:.2f}%.".format(
                    nzero, nzero * 100.0 / rna_data.shape[0]
                )
            )
        hto_small = hashing_data[rna_data.obs_names[idx_df],].X.toarray()

        rna_data.obsm["raw_probs"] = np.zeros((rna_data.shape[0], nsample + 1))
        rna_data.obsm["raw_probs"][:, nsample] = 1.0

        iter_array = [
            (hto_small[i,], hashing_data.uns["background_probs"], alpha, alpha_noise, tol)
            for i in range(ncalc)
        ]
        with multiprocessing.Pool(n_threads) as pool:
            rna_data.obsm["raw_probs"][idx_df, :] = pool.starmap(estimate_probs, iter_array)

        calc_demux(rna_data, hashing_data, nsample, min_signal)

        if has_duplicate_names(hashing_data.var_names):
            rna_data.obs["assignment.dedup"] = rna_data.obs["assignment"]
            rna_data.obs["assignment"] = remove_suffix(rna_data.obs["assignment"].values)

    logger.info("Demultiplexing is done.")


def attach_demux_results(input_rna_file: str, rna_data: UnimodalData) -> MultimodalData:
    """ Write demultiplexing results into raw gene expression matrix.

    Parameters
    ----------
    input_rna_file: ``str``
        Input file for the raw gene count matrix.

    rna_data: ``UnimodalData``
        Processed gene count matrix containing demultiplexing results

    Returns
    -------
    ``MultimodalData``
    A multimodal data object.

    Examples
    --------
    >>> data = attach_demux_results('raw_data.h5', rna_data)
    """
    demux_results = read_input(input_rna_file)
    demux_results.subset_data(modality_subset=['rna'])
    # Assume all matrices are of the same dimension
    assert demux_results.uns["modality"] == "rna"
    barcodes = demux_results.obs_names
    idx = barcodes.isin(rna_data.obs_names)
    selected = barcodes[idx]

    demux_type = np.empty(barcodes.size, dtype="object")
    demux_type[:] = ""
    demux_type[idx] = rna_data.obs.loc[selected, "demux_type"]

    assignment = np.empty(barcodes.size, dtype="object")
    assignment[:] = ""
    assignment[idx] = rna_data.obs.loc[selected, "assignment"]

    assignment_dedup = None
    if "assignment.dedup" in rna_data.obs:
        assignment_dedup = np.empty(barcodes.size, dtype="object")
        assignment_dedup[:] = ""
        assignment_dedup[idx] = rna_data.obs.loc[selected, "assignment.dedup"]

    for keyword in demux_results.list_data():
        unidata = demux_results.get_data(keyword)
        assert unidata.uns["modality"] == "rna"
        unidata.obs["demux_type"] = demux_type
        unidata.obs["assignment"] = assignment
        if assignment_dedup is not None:
            unidata.obs["assignment.dedup"] = assignment_dedup

    logger.info("Demultiplexing results are added to raw expression matrices.")

    return demux_results



from setuptools import setup, find_packages
from setuptools.extension import Extension
from codecs import open
import os


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

requires = [
    "docopt",
    "numpy",
    "pandas",
    "scipy",
    "scikit-learn",
    "seaborn",
    "importlib-metadata >=0.7; python_version < '3.8'",
    "pegasusio >=0.2.12",
]

setup(
    name="demuxEM",
    use_scm_version=True,
    zip_safe=False,
    description="demuxEM is the demultiplexing module of Pegasus",
    long_description=long_description,
    url="https://github.com/klarman-cell-observatory/demuxEM",
    author="Yiming Yang, Joshua Gould and Bo Li",
    author_email="cumulus-support@googlegroups.com",
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="demultiplexing cell/nucleus hashing single-cell data",
    packages=find_packages(),
    install_requires=requires,
    setup_requires = ["setuptools_scm"],
    python_requires="~=3.5",
    entry_points={"console_scripts": ["demuxEM=demuxEM.__main__:main"]},
)
